#include <stdio.h>
#include "mpi.h"



void Laplace_Iter_incre(
                         int           Ndim,
                         int           Glb_Grid_sz[],
                         int           task_Grid_endpts[],
                         int           stride_width[],
                         MPI_Datatype  MD_type,
                         int          *Task_Grid_Modify,

                         void         *Task_Grid
                       )
{

#  if defined( DEBUG ) || defined( DEBUG_stdout )
#  include "debug.c.h"
   int       MD_type_sz;
#  endif

#  include "boolean.c.h"

   int      *task_grid_sz;
   int      *bndry_lgths;
   int      *Nplet_idx;
   int      *Nplet_Ngbr_idx;
   int       hyper_vol;

   int      *Grid_iptr, isum;
   float    *Grid_fptr, fsum0, fsum1;
   double   *Grid_dptr, dsum0, dsum1;

   double    Two_Ndim;
   int       idir, idx, Ngbr_idx, ii;


/*  Allocate Memory for local variables  */

   task_grid_sz      = ( int* ) calloc( Ndim, sizeof( int ) );
   bndry_lgths       = ( int* ) calloc( Ndim, sizeof( int ) );
   Nplet_idx         = ( int* ) calloc( Ndim, sizeof( int ) );
   Nplet_Ngbr_idx    = ( int* ) calloc( Ndim, sizeof( int ) );

/*  Define the boundary lengths of the task grid  */

   hyper_vol = 1;
   for ( idir = 0; idir < Ndim; idir++ )
   {

      task_grid_sz[ idir ] = task_Grid_endpts[ idir + Ndim ]
                           - task_Grid_endpts[ idir ] + 1;
      bndry_lgths[ idir ]  = task_grid_sz[ idir ]
                           + 2 * stride_width[ idir ];
      hyper_vol           *= bndry_lgths[ idir ];
   }

#  if defined( DEBUG ) || defined( DEBUG_stdout )
     ierr_debug = MPI_Comm_rank( MPI_COMM_WORLD, &my_tid_debug );
#  endif

#  if defined( DEBUG ) || defined( DEBUG_stdout )
      MPI_Type_size( MD_type, &MD_type_sz );
      fprintf( stdout, "**** MD_type = %d has %d bytes\n",
               MD_type, MD_type_sz );
#  endif

#  ifdef DEBUG_stdout
      fprintf( stdout, "Task %d's array Task_Grid[] at 1 :\n", my_tid_debug );
      fprintf( stdout, "\n" );
      print_Task_Grid( stdout, Ndim, bndry_lgths, MD_type, Task_Grid );
      fprintf( stdout, "\n" );
#  endif

   switch ( MD_type )
   {
      case MPI_INTEGER:
         Grid_iptr = (int*) Task_Grid;
         break;
      case MPI_REAL:
         Grid_fptr = (float*) Task_Grid;
         break;
      case MPI_DOUBLE_PRECISION:
         Grid_dptr = (double*) Task_Grid;
         break;
      case MPI_COMPLEX:
         Grid_fptr = (float*) Task_Grid;
         break;
      case MPI_DOUBLE_COMPLEX:
         Grid_dptr = (double*) Task_Grid;
         break;
      default:
         fprintf( stderr, __FILE__": Wrong type in laplace_iter()\n" );
         break;
   }

/*  Update Task_Grid[]   */

   switch ( MD_type )
   {

      case MPI_INTEGER:
         Two_Ndim = 2.0 * Ndim;
         for ( idx = 0; idx < hyper_vol; idx++ )
         {
            if ( Task_Grid_Modify[ idx ] == iTRUE )
            {
               adr_lnr_2_Nplet( idx, Ndim, bndry_lgths, Nplet_idx );

               isum = 0;
               for ( idir = 0; idir < Ndim; idir++ )
               {
                  for ( ii = 0; ii < Ndim; ii++ )
                     Nplet_Ngbr_idx[ ii ] = Nplet_idx[ ii ];

                  Nplet_Ngbr_idx[ idir ] += 1;
                  adr_Nplet_2_lnr( Ndim, bndry_lgths, Nplet_Ngbr_idx,
                                   &Ngbr_idx );
                  isum += Grid_iptr[ Ngbr_idx ];

                  Nplet_Ngbr_idx[ idir ] -= 2;
                  adr_Nplet_2_lnr( Ndim, bndry_lgths, Nplet_Ngbr_idx,
                                   &Ngbr_idx );
                  isum += Grid_iptr[ Ngbr_idx ];
               }
               isum /= Two_Ndim;

               Grid_iptr[ idx ] = isum;
            }   /*   if ( Task_Grid_Modify[ idx ] == iTRUE )   */
         }      /*   for ( idx = 0; idx < hyper_vol; idx++ )   */
         break;

      case MPI_REAL:
         Two_Ndim = 2.0 * Ndim;
         for ( idx = 0; idx < hyper_vol; idx++ )
         {
            if ( Task_Grid_Modify[ idx ] == iTRUE )
            {
               adr_lnr_2_Nplet( idx, Ndim, bndry_lgths, Nplet_idx );

               fsum0 = 0.0;
               for ( idir = 0; idir < Ndim; idir++ )
               {
                  for ( ii = 0; ii < Ndim; ii++ )
                     Nplet_Ngbr_idx[ ii ] = Nplet_idx[ ii ];

                  Nplet_Ngbr_idx[ idir ] += 1;
                  adr_Nplet_2_lnr( Ndim, bndry_lgths, Nplet_Ngbr_idx,
                                   &Ngbr_idx );
                  fsum0 += Grid_fptr[ Ngbr_idx ];

                  Nplet_Ngbr_idx[ idir ] -= 2;
                  adr_Nplet_2_lnr( Ndim, bndry_lgths, Nplet_Ngbr_idx,
                                   &Ngbr_idx );
                  fsum0 += Grid_fptr[ Ngbr_idx ];
               }
               fsum0 /= Two_Ndim;

               Grid_iptr[ idx ] = fsum0;
            }   /*   if ( Task_Grid_Modify[ idx ] == iTRUE )   */
         }      /*   for ( idx = 0; idx < hyper_vol; idx++ )   */
         break;

      case MPI_DOUBLE_PRECISION:
         Two_Ndim = 2.0 * Ndim;
         for ( idx = 0; idx < hyper_vol; idx++ )
         {
            if ( Task_Grid_Modify[ idx ] == iTRUE )
            {
               adr_lnr_2_Nplet( idx, Ndim, bndry_lgths, Nplet_idx );
         
               dsum0 = 0.0;
               for ( idir = 0; idir < Ndim; idir++ )
               {
                  for ( ii = 0; ii < Ndim; ii++ )
                     Nplet_Ngbr_idx[ ii ] = Nplet_idx[ ii ];

                  Nplet_Ngbr_idx[ idir ] += 1;
                  adr_Nplet_2_lnr( Ndim, bndry_lgths, Nplet_Ngbr_idx,
                                   &Ngbr_idx );
                  dsum0 += Grid_dptr[ Ngbr_idx ];

                  Nplet_Ngbr_idx[ idir ] -= 2;
                  adr_Nplet_2_lnr( Ndim, bndry_lgths, Nplet_Ngbr_idx,
                                   &Ngbr_idx );
                  dsum0 += Grid_dptr[ Ngbr_idx ];
               }
               dsum0 /= Two_Ndim;

               Grid_dptr[ idx ] = dsum0;
            }   /*   if ( Task_Grid_Modify[ idx ] == iTRUE )   */
         }      /*   for ( idx = 0; idx < hyper_vol; idx++ )   */
         break;

      case MPI_COMPLEX:
         Two_Ndim = 2.0 * Ndim;
         for ( idx = 0; idx < hyper_vol; idx++ )
         {
            if ( Task_Grid_Modify[ idx ] == iTRUE )
            {
               adr_lnr_2_Nplet( idx, Ndim, bndry_lgths, Nplet_idx );

               fsum0 = 0.0;
               fsum1 = 0.0;
               for ( idir = 0; idir < Ndim; idir++ )
               {
                  for ( ii = 0; ii < Ndim; ii++ )
                     Nplet_Ngbr_idx[ ii ] = Nplet_idx[ ii ];

                  Nplet_Ngbr_idx[ idir ] += 1;
                  adr_Nplet_2_lnr( Ndim, bndry_lgths, Nplet_Ngbr_idx,
                                   &Ngbr_idx );
                  fsum0 += Grid_fptr[ 2*Ngbr_idx ];
                  fsum1 += Grid_fptr[ 2*Ngbr_idx+1 ];

                  Nplet_Ngbr_idx[ idir ] -= 2;
                  adr_Nplet_2_lnr( Ndim, bndry_lgths, Nplet_Ngbr_idx,
                                   &Ngbr_idx );
                  fsum0 += Grid_fptr[ 2*Ngbr_idx ];
                  fsum1 += Grid_fptr[ 2*Ngbr_idx+1 ];
               }
               fsum0 /= Two_Ndim;
               fsum1 /= Two_Ndim;

               Grid_iptr[ 2*idx   ] = fsum0;
               Grid_iptr[ 2*idx+1 ] = fsum1;
            }   /*   if ( Task_Grid_Modify[ idx ] == iTRUE )   */
         }      /*   for ( idx = 0; idx < hyper_vol; idx++ )   */
         break;

      case MPI_DOUBLE_COMPLEX:
         Two_Ndim = 2.0 * Ndim;
         for ( idx = 0; idx < hyper_vol; idx++ )
         {
            if ( Task_Grid_Modify[ idx ] == iTRUE )
            {
               adr_lnr_2_Nplet( idx, Ndim, bndry_lgths, Nplet_idx );

               dsum0 = 0.0;
               dsum1 = 0.0;
               for ( idir = 0; idir < Ndim; idir++ )
               {
                  for ( ii = 0; ii < Ndim; ii++ )
                     Nplet_Ngbr_idx[ ii ] = Nplet_idx[ ii ];

                  Nplet_Ngbr_idx[ idir ] += 1;
                  adr_Nplet_2_lnr( Ndim, bndry_lgths, Nplet_Ngbr_idx,
                                   &Ngbr_idx );
                  dsum0 += Grid_dptr[ 2*Ngbr_idx ];
                  dsum1 += Grid_dptr[ 2*Ngbr_idx+1 ];

                  Nplet_Ngbr_idx[ idir ] -= 2;
                  adr_Nplet_2_lnr( Ndim, bndry_lgths, Nplet_Ngbr_idx,
                                   &Ngbr_idx );
                  dsum0 += Grid_dptr[ 2*Ngbr_idx ];
                  dsum1 += Grid_dptr[ 2*Ngbr_idx+1 ];
               }
               dsum0 /= Two_Ndim; 
               dsum1 /= Two_Ndim;

               Grid_iptr[ 2*idx   ] = dsum0;
               Grid_iptr[ 2*idx+1 ] = dsum1;
            }   /*   if ( Task_Grid_Modify[ idx ] == iTRUE )   */
         }      /*   for ( idx = 0; idx < hyper_vol; idx++ )   */
         break;

   }            /*   switch ( MD_type )   */

#  ifdef DEBUG
      sprintf( filename_debug, filename_debug_fmt, my_tid_debug );
      file_debug = fopen( filename_debug, "a+" );
      fprintf( file_debug, "\n" );
      print_Task_Grid( file_debug, Ndim, bndry_lgths, MD_type, Task_Grid );
      fprintf( file_debug, "\n" );
      fclose( file_debug );

      fprintf( stdout, "Task %d's array Task_Grid[] at 2 :\n", my_tid_debug );
      fprintf( stdout, "\n" );
      print_Task_Grid( stdout, Ndim, bndry_lgths, MD_type, Task_Grid );
      fprintf( stdout, "\n" );
#  endif

   free( Nplet_Ngbr_idx );
   free( Nplet_idx );
   free( bndry_lgths );
   free( task_grid_sz );

}



#ifdef FortranSingleUnderscore
#   define laplace_iter_incre__ laplace_iter_incre__
#endif

/*  Fortran interface for Laplace_Iter_incre()  */
void laplace_iter_incre__( 
                     int          *Ndim,
                     int          *Glb_Grid_sz,
                     int          *task_Grid_endpts,
                     int          *stride_width,
                     MPI_Datatype *MD_type,
                     int          *Task_Grid_Modify,

                     void         *Task_Grid
                   )
{
   Laplace_Iter_incre( *Ndim, Glb_Grid_sz, task_Grid_endpts, stride_width,
                       *MD_type, Task_Grid_Modify, Task_Grid );  
}
