#include <stdio.h>
#if defined( sgi )
#include <alloca.h>
#else
#include <stdlib.h>
#endif

#include "mpi.h"



void last_ctg_dim(
                   int  Ndim,
                   int  bndry_lgths[], 
                   int  Nplet[],

                   int *last_matched_ctg_idx
                 )
{
   /*  local variables  */
   int matched_flag = 1;
   int idx = 0;

   while ( idx < Ndim && matched_flag ) 
   {
      if ( Nplet[ idx ] != bndry_lgths[ idx ]-1 ) matched_flag = 0;
      idx += 1;
   }

   *last_matched_ctg_idx = idx - 2;
}



/*@C
   print_Task_Grid - print the multi-dimensional continuous block data,
                     Grid_data[], according to the given size in each dimension
                     ii, bndry_lgths[ ii ].  bndry_lgths[ ii=0 ] corresponds
                     to the maximum size of the fastest changing index of
                     array Grid_data[].  The data is printed out according
                     to their memory layout and is splitted into the form of
                     2-dimensional table. The fastest changing index
                     becomes table's column index and the second fastest
                     changing index becomes table's row index.  The rest of
                     the indices are used to label all the 2-dimensional
                     tables printed. 

  Unmodified Input Variables :
. fout - File Pointer to the output.
. Ndim - Dimensionality of the multi-dimensional block Array Grid_data[].
. bndry_lgths - Integer Array of size Ndim.  Each element, bndry_lgths[ ii ],
                is the maximum size in the ii-th dimension.
. MD_type - MPI_Datatype of the block data, Grid_data[].
. Grid_data - Void Pointer to the continuous block data.
.n

  Notes on the Fortran routines :
  The Fortran counterparts of this C routines are called
  print_Task_xxGrid( ch_prt, Ndim, bdy_lgths, Grid_data ) with Ndim,
  bdy_lgths[], Grid_data[] defined similarly as in C routines, except 
  ch_prt is an integer for the fortran file channel number.  Also for
  different datatype of Grid_data[], xx in the subroutine name will be
  different.
.vb
  integer      array Grid_data:  xx=i
  real*4       array Grid_data:  xx=r4
  real*8       array Grid_data:  xx=r8
  complex*8    array Grid_data:  xx=c4
  complex*16   array grid_data:  xx=c8
.ve

.N RGX_not_included

@*/

void print_Task_Grid(
                      FILE         *fout,
                      int           Ndim,
                      int           bndry_lgths[],
                      MPI_Datatype  MD_type,
                      void         *Grid_data
                    )
{
    /*  Local variables: */
    int    *Nplet;
    int     row_lgth;
    int     hyper_vol;
    int     elem_idx;
    int     last_matched_ctg_idx;
    char    format_str[5];
    int    *Grid_iptr;
    float  *Grid_fptr;
    double *Grid_dptr;
    int   ii;



    Nplet = ( int * ) alloca( Ndim * sizeof( int ) );

    hyper_vol = 1;
    for ( ii = 0; ii < Ndim; ii++ )
    {
       hyper_vol *= bndry_lgths[ ii ];
       Nplet[ ii ] = 0;
    }

    row_lgth = bndry_lgths[ 0 ];

    switch ( MD_type )
    {
       case MPI_INTEGER:
          elem_idx = 0;
          Grid_iptr = (int*)Grid_data;
          while ( elem_idx < hyper_vol )
          {
             for ( ii = elem_idx; ii < elem_idx+row_lgth; ii++ )
                fprintf( fout, "%d ", Grid_iptr[ ii ] );
             fprintf( fout, "\n" );

             elem_idx += row_lgth;
             adr_lnr_2_Nplet( elem_idx, Ndim, bndry_lgths, Nplet );
             last_ctg_dim( Ndim, bndry_lgths, Nplet, &last_matched_ctg_idx );
             for ( ii = 0; ii < last_matched_ctg_idx; ii++ )
             fprintf( fout, "-------------------------------------\n" );
          }
          break;
       case MPI_REAL:
          elem_idx = 0;
          Grid_fptr = (float*)Grid_data;
          while ( elem_idx < hyper_vol )
          {
             for ( ii = elem_idx; ii < elem_idx+row_lgth; ii++ )
                fprintf( fout, "%f ", Grid_fptr[ ii ] );
             fprintf( fout, "\n" );

             elem_idx += row_lgth;
             adr_lnr_2_Nplet( elem_idx, Ndim, bndry_lgths, Nplet );
             last_ctg_dim( Ndim, bndry_lgths, Nplet, &last_matched_ctg_idx );
             for ( ii = 0; ii < last_matched_ctg_idx; ii++ )
             fprintf( fout, "-------------------------------------\n" );
          }
          break;
       case MPI_DOUBLE_PRECISION:
          elem_idx = 0;
          Grid_dptr = (double*)Grid_data;
          while ( elem_idx < hyper_vol )
          {
             for ( ii = elem_idx; ii < elem_idx+row_lgth; ii++ )
                fprintf( fout, "%f ", Grid_dptr[ ii ] );
             fprintf( fout, "\n" );

             elem_idx += row_lgth;
             adr_lnr_2_Nplet( elem_idx, Ndim, bndry_lgths, Nplet );
             last_ctg_dim( Ndim, bndry_lgths, Nplet, &last_matched_ctg_idx );
             for ( ii = 0; ii < last_matched_ctg_idx; ii++ )
             fprintf( fout, "-------------------------------------\n" );
          }
          break;
       case MPI_COMPLEX:
          elem_idx = 0;
          Grid_fptr = (float*)Grid_data;
          while ( elem_idx < hyper_vol )
          {
             for ( ii = elem_idx; ii < elem_idx+row_lgth; ii++ )
                fprintf( fout, "(%f,%f) ", Grid_fptr[2*ii], Grid_fptr[2*ii+1] );
             fprintf( fout, "\n" );

             elem_idx += row_lgth;
             adr_lnr_2_Nplet( elem_idx, Ndim, bndry_lgths, Nplet );
             last_ctg_dim( Ndim, bndry_lgths, Nplet, &last_matched_ctg_idx );
             for ( ii = 0; ii < last_matched_ctg_idx; ii++ )
             fprintf( fout, "-------------------------------------\n" );
          }
          break;
       case MPI_DOUBLE_COMPLEX:
          elem_idx = 0;
          Grid_dptr = (double*)Grid_data;
          while ( elem_idx < hyper_vol )
          {
             for ( ii = elem_idx; ii < elem_idx+row_lgth; ii++ )
                fprintf( fout, "(%f,%f) ", Grid_dptr[2*ii], Grid_dptr[2*ii+1] );
             fprintf( fout, "\n" );

             elem_idx += row_lgth;
             adr_lnr_2_Nplet( elem_idx, Ndim, bndry_lgths, Nplet );
             last_ctg_dim( Ndim, bndry_lgths, Nplet, &last_matched_ctg_idx );
             for ( ii = 0; ii < last_matched_ctg_idx; ii++ )
             fprintf( fout, "-------------------------------------\n" );
          }
          break;
       default:
          fprintf( stderr, __FILE__": NO matching data type for type %d\n",
                   MD_type );
          return;
          break;
    }

}
